'''
Module that contains the bootraping tests to evaluate the performance of the fuzzy rules and fuzzy rule based classifiers.


'''

import numpy as np
import pandas as pd

from sklearn.base import ClassifierMixin

try:
    from . import rules
    from . import fuzzy_sets as fs
except:
    import rules
    import fuzzy_sets as fs


def generate_bootstrap_samples(X: np.array, y: np.array, n_samples:int):
    '''
    Generates bootstrap samples from the given data.

    :param X: data to sample.
    :param y: labels to sample.
    :param n_samples: number of samples to generate.
    :return: a list of tuples with the sampled data and labels.
    '''
    samples = []
    for i in range(n_samples):
        idx = np.random.choice(X.shape[0], X.shape[0], replace=True)
        samples.append((X[idx], y[idx]))
    
    return samples


def generate_random_rules(nAnts: int, nRules: int, n_linguistic_variables: int, class_names: np.array):
    '''
    Generates random rules for the given parameters.

    :param nAnts: number of ants in the classifier.
    :param nRules: number of rules in the classifier.
    :param n_linguistic_variables: number of linguistic variables in the classifier.
    :param class_names: classes in the classifier.
    :return: a list of tuples with the antecedents and consequents of the rules.
    '''
    rules = []
    for i in range(nRules):
        antecedents = []
        for j in range(nAnts):
            antecedents.append(np.random.randint(-1, n_linguistic_variables, size=1)[0])
        
        consequent = np.random.choice(class_names)
        rules.append((antecedents, consequent))
    
    return rules


def membership_randomrule(rule, X: np.array, fuzzy_variables: list[fs.fuzzyVariable]):
    '''
    Computes the membership of the given data to the rule.

    :param rule: rule to evaluate.
    :param X: data to evaluate.
    :param fuzzy_variables: fuzzy variables of the classifier.
    :return: the membership of the data to the rule.
    '''
    if fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t1:
        membership = np.zeros(X.shape[0]) + 1
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t2:
        membership = np.zeros((X.shape[0], 2)) + 1
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
        membership = np.zeros((X.shape[0], len(fuzzy_variables[0][0].alpha_cuts), 2)) + 1

    else:
        raise ValueError('Fuzzy type not supported')
    
    for i in range(len(rule[0])):
        if rule[0][i] != -1:
            membership = membership * fuzzy_variables[i][rule[0][i]].membership(X[:, i])
    
    return membership


def quality_metric_rule(rule, X: np.array, y: np.array, fuzzy_variables: list[fs.fuzzyVariable]):
    '''
    Computes the quality of the rule as the percentage of samples of the consequent class that have more membership to the rule than to the other classes.

    :param rule: rule to evaluate.
    :param X: data to evaluate.
    :param y: labels to evaluate.
    :param fuzzy_variables: fuzzy variables of the classifier.
    :param class_names: classes in the classifier.
    :return: the cross entropy of the rule.
    '''
    membership = membership_randomrule(rule, X, fuzzy_variables)
    quality_metric = 0
    class_names = np.unique(y)

    class_samples = np.where(y == rule[1])[0]
    if fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t1:
        membership_class = np.mean(membership[class_samples])
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t2:
        membership_class = np.mean(membership[class_samples], axis=0).mean()
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
        membership_class = fuzzy_variables[0][0].alpha_reduction(membership[class_samples])
        membership_class = membership_class.mean(axis=1).mean()

    non_class_samples = np.where(y != rule[1])[0]
    if fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t1:
        membership_non_class = np.mean(membership[non_class_samples])
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.t2:
        membership_non_class = np.mean(membership[non_class_samples], axis=0).mean()
    elif fuzzy_variables[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
        membership_non_class = fuzzy_variables[0][0].alpha_reduction(membership[non_class_samples])
        membership_non_class = membership_non_class.mean(axis=1).mean()


    quality_metric = membership_class - membership_non_class

    return quality_metric 


def generate_null_distribution(classifier, X: np.array, y: np.array, n_samples:int, nRules: int):
    '''
    Generates the null distribution of the classifier.

    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param y: labels to evaluate.
    :param n_samples: number of samples to generate.
    :param nRules: number of rules to bootstrap.
    :return: a list with the cross entropy of the rules in the null distribution.
    '''
    null_distribution = []
    nAnts = X.shape[1]
    class_names = np.unique(y)
    lvs = [fv for fv in classifier.lvs] if isinstance(classifier, ClassifierMixin) else [fv for fv in classifier[0].antecedents]
    
    n_linguistic_variables = [len(fv) for fv in lvs]
    
    for i in range(n_samples):
        rules = generate_random_rules(nAnts, nRules, n_linguistic_variables, class_names)
        qualities_metric = []
        for rule in rules:
            qualities_metric.append(quality_metric_rule(rule, X, y, lvs))
        null_distribution.append(qualities_metric)
    
    return np.array(null_distribution)


def _aux_compute_rule_p_value(rule, X: np.array, y: np.array, fuzzy_variables: list[fs.fuzzyVariable], null_distribution: np.array):
    '''
    Computes the p-value of the rule.

    :param rule: rule to evaluate.
    :param X: data to evaluate.
    :param y: labels to evaluate.
    :param fuzzy_variables: fuzzy variables of the classifier.
    :param null_distribution: null distribution of the classifier.
    :return: the p-value of the rule.
    '''
    quality_metric = quality_metric_rule(rule, X, y, fuzzy_variables)
    # Select randomly one column per row
    chosen_columns = np.random.randint(0, null_distribution.shape[1], size=(null_distribution.shape[0], 1))
    chosen_population = null_distribution[np.arange(null_distribution.shape[0]), chosen_columns.flatten()]
    p_value = np.mean(quality_metric < chosen_population)

    return p_value


def compute_rule_p_value(classifier, X: np.array, y: np.array, nSamples: int=100, nRules: int=10):
    '''
    Computes the p-value of the rules in the classifier.

    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param y: labels to evaluate.
    :return: a list with the p-value of the rules.
    '''
    null_distribution = generate_null_distribution(classifier, X, y, nSamples, nRules)
    p_values = []

    if isinstance(classifier, ClassifierMixin):
        crules = classifier.rule_base.get_rules()
        lvs = classifier.lvs
        rconsequents = classifier.rule_base.get_consequents()
    else:
        crules = classifier.get_rules()
        lvs = classifier[0].antecedents
        rconsequents = classifier.get_consequents()
    
    for ix, rule in enumerate(crules):
        p_values.append(_aux_compute_rule_p_value([rule, rconsequents[ix]], X, y, lvs, null_distribution))
    
    return np.array(p_values)



'''
Module to implement different boostrapping tests to evaluate the performance of the fuzzy rules and fuzzy rule based classifiers.

@authors:
    - Javier Fumanal Idocin
    
@references:
    - Permutation Tests for studying Classifier Performance, Ojala, J., Garriga, G. (2010) JMLR 
'''

import numpy as np
import pandas as pd

from sklearn.base import ClassifierMixin

try:
    from . import rules
except:
    import rules

class classifierWrapper(ClassifierMixin):

    def __init__(self, classifier: rules.MasterRuleBase, nrule:int):
        '''
        Creates a classifier wrapper for the given rule in a rule-based classifier.

        :param classifier: rule-based classifier to evaluate.
        :param nrule: rule to evaluate.
        '''
        self.classifier = classifier
        self.nrule = nrule
    

    def predict(self, X: np.array):
        '''
        Predicts the labels of the given data using the rule-based classifier.

        :param X: data to evaluate.
        :return: predicted labels.
        '''
        winning_rule_preds, _winning_association_degrees = self.classifier._winning_rules(X)

        return winning_rule_preds == self.nrule


def _aux_ova_rule_classifier(classifier: rules.MasterRuleBase, nrule:int, labels: np.array):
    '''
    Formats the data to the OVA format and creates a classifier wrapper for the given rule in a rule-based classifier.

    :param classifier: rule-based classifier to evaluate.
    :param nrule: rule to evaluate.
    :param labels: labels to evaluate.
    :return: an object of the sklearn classifier class.
    '''
    ruleOVA_classifier = classifierWrapper(classifier, nrule)
    rules_consequents = classifier.get_consequents()

    new_labels = labels == rules_consequents[nrule]

    return ruleOVA_classifier, new_labels
    

def permutation_p_compute(error_rates: np.array, og_error_rate: float):
    """
    Function to compute the p-value of a permutation test.

    :param true_values: list of true values of the test.
    :param error_rates: list of error rates of the test.
    :param og_error_rate: original error rate of the test.
    :return: p-value of the permutation test.
    """
    sum_random_victories = error_rates < og_error_rate
    p_value = (np.sum(sum_random_victories)+1) / (len(error_rates)+1)

    return p_value


def estimate_og_error_rate(classifier, X: np.array, labels: np.array, r:int=10):
    '''
    Function to estimate the original error rate of a classifier.
    
    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param labels: labels to evaluate.
    :return: original error rate of the classifier.
    '''
    og_error_rates = []
    for _ in range(r): 
        og_error_rate0 = np.mean(classifier.predict(X) != labels)
        og_error_rates.append(og_error_rate0)

    og_error_rate = np.mean(og_error_rates)

    return og_error_rate


def permutation_labels_test(classifier, X, labels, k=100, r=10):
    '''
    Function to perform a permutation test to evaluate the performance of a classifier.
    
    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param labels: labels to evaluate.
    :param k: number of permutations to perform.
    :param r: number of repetitions to estimate the original error rate.
    :return: p-value of the permutation test.
    '''
    error_rates = []
    og_error_rate = estimate_og_error_rate(classifier, X, labels, r=r)
    og_predictions = classifier.predict(X)

    for i in range(k):
        permuted_labels = np.random.permutation(labels)
        permuted_error_rate = np.mean(permuted_labels != og_predictions)
        error_rates.append(permuted_error_rate)

    
    p_value = permutation_p_compute(np.array(error_rates), og_error_rate)

    return p_value


def permute_columns_class_test(classifier, X:np.array, labels:np.array, k:int=100, r:int=10):
    '''
    Function to perform a permutation test to evaluate the performance of a classifier.
    
    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param labels: labels to evaluate.
    :param k: number of permutations to perform.
    :param r: number of repetitions to estimate the original error rate.
    :return: p-value of the permutation test.
    '''
    error_rates = []
    og_error_rate = estimate_og_error_rate(classifier, X, labels, r=r)
    classes = np.unique(labels)

    for i in range(k):
        permuted_X = X.copy()
        for clas in classes:
            premuted_x_clas = permuted_X[labels == clas]
            premuted_x_clas = np.random.permutation(premuted_x_clas)
            permuted_X[labels == clas] = premuted_x_clas
            
        permuted_error_rate = np.mean(classifier.predict(permuted_X) != labels)
        error_rates.append(permuted_error_rate)

    p_value = permutation_p_compute(np.array(error_rates), og_error_rate)

    return p_value


def rulewise_label_permutation_test(classifier: rules.MasterRuleBase, X: np.array, labels: np.array, k:int=100, r:int=10):
    '''
    Function to perform a permutation test to evaluate the performance of a classifier.
    
    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param labels: labels to evaluate.
    :param k: number of permutations to perform.
    :param r: number of repetitions to estimate the original error rate.
    :return: p-value of the permutation test.
    '''
    for ix_rule, rule in enumerate(classifier.get_rules()):
        error_rates = []

        ruleOVA_classifier, new_labels = _aux_ova_rule_classifier(classifier, ix_rule, labels)
        og_error_rate = estimate_og_error_rate(ruleOVA_classifier, X, new_labels, r=r)
        og_predictions = ruleOVA_classifier.predict(X)

        for i in range(k):
            permuted_labels = np.random.permutation(new_labels)
            permuted_error_rate = np.mean(permuted_labels != og_predictions)
            error_rates.append(permuted_error_rate)

    
        p_value = permutation_p_compute(np.array(error_rates), og_error_rate)
        rule.p_value_class_structure = p_value


def rulewise_column_permutation_test(classifier: rules.MasterRuleBase, X: np.array, labels: np.array, k:int=100, r:int=10):
    '''
    Function to perform a permutation test to evaluate the performance of a classifier.
    
    :param classifier: classifier to evaluate.
    :param X: data to evaluate.
    :param labels: labels to evaluate.
    :param k: number of permutations to perform.
    :param r: number of repetitions to estimate the original error rate.
    :return: p-value of the permutation test.
    '''
    for ix_rule, rule in enumerate(classifier.get_rules()):
        error_rates = []

        ruleOVA_classifier, new_labels = _aux_ova_rule_classifier(classifier, ix_rule, labels)
        og_error_rate = estimate_og_error_rate(ruleOVA_classifier, X, new_labels, r=r)
        classes = np.unique(labels)

        for i in range(k):
            permuted_X = X.copy()
            clas = classifier.get_consequents()[ix_rule]
            premuted_x_clas = permuted_X[new_labels == clas]
            premuted_x_clas = np.random.permutation(premuted_x_clas)
            permuted_X[new_labels == clas] = premuted_x_clas
                
            permuted_error_rate = np.mean(ruleOVA_classifier.predict(permuted_X) != new_labels)
            error_rates.append(permuted_error_rate)

        p_value = permutation_p_compute(np.array(error_rates), og_error_rate)
        rule.p_value_feature_coalitions = p_value

    


'''
A module to analyze the stability of the patterns generated by the fuzzy classifier. It runs the same experiment n times and then it counts the number of unique patterns that appear in the rule base and how often they do. It also counts the number of times each variable is used in the rule base.
'''
import numpy as np
from multiprocessing.pool import ThreadPool
from pymoo.core.problem import StarmapParallelization
from sklearn.model_selection import train_test_split
import numbers
import matplotlib.pyplot as plt
from matplotlib import colormaps
import random

try:
    from . import fuzzy_sets as fs
    from . import maintenance as mnt
    from . import rules as rl
    from . import evolutionary_fit as evf
    from . import vis_rules
    from . import eval_rules as evr
    
except:
    import fuzzy_sets as fs
    import maintenance as mnt
    import rules as rl
    import evolutionary_fit as evf
    import vis_rules
    import eval_rules as evr


def add_dicts(dict1: dict, dict2: dict):
    # We will add the values of dict2 to dict1
    for key in dict2:
        try:
            dict1[key] += dict2[key]
        except KeyError:
            dict1[key] = dict2[key]
    
    return dict1

def concatenate_dicts(dict1: dict, dict2: dict):
    # We will concatenate the values of dict2 to dict1
    for key in dict2:
        try:
            dict1[key]
        except KeyError:
            dict1[key] = dict2[key]
    
    return dict1


def str_rule_as_list(rule:str):
    # We will transform the string form of a rule into a list of integers
    rule = rule.replace('[', '')
    rule = rule.replace(']', '')
    rule = rule.replace('(', '')
    rule = rule.replace(')', '')
    rule = rule.replace('.', '')
    
    rule = rule.split()

    return [int(rr) for rr in rule]


class pattern_stabilizer():

    def __init__(self,  X, y, nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = fs.FUZZY_SETS.t1, tolerance: float = 0.0, class_names: list[str] = None,
                 n_linguistic_variables: int = 3, verbose=False, linguistic_variables: list[fs.fuzzyVariable] = None,
                 domain: list[float] = None, n_class: int=None, runner: int=1, ds_mode:int=0, allow_unkown:bool=False,
                 fuzzy_modifiers:bool=False, stratify_by:str=None) -> None:
        
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the dominance score of the rules.
        :param n_linguist_variables: number of linguistic variables per antecedent.
        :param verbose: if True, prints the progress of the optimization.
        :param linguistic_variables: list of fuzzyVariables type. If None (default) the optimization process will init+optimize them.
        :param domain: list of the limits for each variable. If None (default) the classifier will compute them empirically.
        :param n_class: names of the classes in the problem. If None (default) the classifier will compute it empirically.
        :param precomputed_rules: MasterRuleBase object. If not None, the classifier will use the rules in the object and ignore the conflicting parameters.
        :param runner: number of threads to use. If None (default) the classifier will use 1 thread.
        :param stratify_by: string naming additional column to stratify test/train split if desired (e.g. split by participant rather than trial).
        '''
        self.nRules = nRules
        self.nAnts = nAnts
        self.nclasses_ = n_class

        if class_names is None:
            self.classes_names = list(np.unique(y))
        else:
            if isinstance(class_names, np.ndarray):
                self.classes_names = list(class_names)
            else:
                self.classes_names = class_names

        self.custom_loss = None
        self.verbose = verbose
        self.tolerance = tolerance
        self.ds_mode = ds_mode
        self.allow_unknown = allow_unkown
        self.fuzzy_modifiers = fuzzy_modifiers
        self.runner = runner
        
        if linguistic_variables is not None:
            # If the linguistic variables are precomputed then we act accordingly
            self.lvs = linguistic_variables
            self.n_linguist_variables = [len(lv.linguistic_variable_names()) for lv in self.lvs]
            self.domain = None
            self.fuzzy_type = self.lvs[0].fuzzy_type()
        else:
            # If not, then we need the parameters sumistered by the user.
            self.lvs = None
            self.fuzzy_type = fuzzy_type
            self.n_linguist_variables = n_linguistic_variables
            self.domain = domain

        self.alpha_ = 0.0
        self.beta_ = 0.0

        self.X = X
        self.y = y

        self.stratify_by = stratify_by


    def generate_solutions(self, n=30, n_gen=10, pop_size=10,stratify_by:str=None,test_size=0.33):
        # We will generate n solutions and return the rule bases and the accuracies
        # Pre-defined stratification by additional column (e.g. participant vs. trial) if desired
        rule_bases = []
        accs = []
        use_names = not isinstance(self.classes_names[0], numbers.Number)
        strat_column = None
        strat = self.stratify_by

        for ix in range(n):
            fl_classifier = evf.BaseFuzzyRulesClassifier(nRules=self.nRules, linguistic_variables=self.lvs, nAnts=self.nAnts, n_linguistic_variables=self.n_linguist_variables, fuzzy_type=self.fuzzy_type, verbose=False, tolerance=self.tolerance, runner=self.runner, ds_mode=self.ds_mode, fuzzy_modifiers=self.fuzzy_modifiers, allow_unknown=self.allow_unknown)
            
            if strat:
                strat_column = self.stratify_by
                # Use strat_column to split the data then drop the column and perform test/train split
                id = self.X[strat_column].unique() #get unique ids in strat_column
                random.seed(ix) # set different seeds to split data differently per iteration
                test_ix = random.sample(range(0,len(id)),int(len(id)*test_size)) #randomly select indices for testing using test_size value
                test_id = [id[i] for i in test_ix] #get labels for test ix

                # Get train/test X and y data
                X_train = self.X[~self.X[strat_column].isin(test_id)]
                X_test = self.X[self.X[strat_column].isin(test_id)]
                y_train = [self.y[i] for i in X_train.index.tolist()]
                y_test = [self.y[i] for i in X_test.index.tolist()]

                # Drop ID column for training
                X_train = X_train.drop(columns=[strat_column])
                X_test = X_test.drop(columns=[strat_column])

                fl_classifier.fit(X_train, np.array(y_train), n_gen=n_gen, pop_size=pop_size, checkpoints=0)
            
            else:
                X_train, X_test, y_train, y_test = train_test_split(self.X, self.y, test_size=test_size, random_state=ix)
                fl_classifier.fit(X_train, np.array(y_train), n_gen=n_gen, pop_size=pop_size, checkpoints=0)


            rule_bases.append(fl_classifier.rule_base)
            accuracy = np.mean(np.equal(fl_classifier.forward(X_test, out_class_names=use_names), np.array(y_test)))
            accs.append(accuracy)
        
        return rule_bases, accs
    
    def count_unique_patterns(self, rule_base: rl.RuleBase):
        '''
        We will count the number of unique patterns in the rule base. It will also count the number of times each variable is used in the rule base and the record the dominance score of each pattern.

        :param rule_base: RuleBase object. The rule base to analyze.
        :return unique_patterns: dict. The dictionary with the unique patterns and the number of times they appear.
        :return patterns_ds: dict. The dictionary with the dominance score of each pattern.
        :return var_used: dict. The dictionary with the number of times each variable is used in the rule base.
        '''
        unique_patterns = {}
        patterns_ds = {}
        var_used = {}

        for ix, rule in enumerate(rule_base.get_rulebase_matrix()):
            pattern = str(rule)
            patterns_ds[pattern] = rule_base[ix].score
            if pattern in unique_patterns:
                unique_patterns[pattern] += 1
            else:
                unique_patterns[pattern] = 1

            for jx, var in enumerate(rule):
                try:
                    var_used[jx][var] += 1
                except:
                    try:
                        var_used[jx][var] = 1
                    except:
                        var_used[jx] = {}
                        var_used[jx][var] = 1
        

        
        return unique_patterns, patterns_ds, var_used
    
    def count_unique_patterns_all_classes(self, mrule_base: rl.MasterRuleBase, class_patterns: dict[list] = None, patterns_dss: dict[list] = None, class_vars: dict[list] = None):
        '''
        Counts the number of unique patterns for all classes. It also counts the number of times each variable is used in the rule base and the dominance score of each pattern.

        :param mrule_base: MasterRuleBase object. The rule base to analyze.
        :param class_patterns: dict[list]. The dictionary with the unique patterns for each class. If None, it will be initialized.
        :param patterns_dss: dict[list]. The dictionary with the dominance score of each pattern for each class. If None, it will be initialized.
        :param class_vars: dict[list]. The dictionary with the number of times each variable is used in the rule base for each class. If None, it will be initialized.
        :return class_patterns: dict[list]. The dictionary with the unique patterns for each class.
        :return patterns_dss: dict[list]. The dictionary with the dominance score of each pattern for each class.
        :return class_vars: dict[list]. The dictionary with the number of times each variable is used in the rule base for each class.
        '''
        if class_patterns is None:
            class_patterns = {ix: {} for ix in range(len(mrule_base))}
            class_vars = {}
            for key in range(len(mrule_base)):
                class_vars[key] = {}
                for jx in range(len(mrule_base.n_linguistic_variables())):
                    class_vars[key][jx] = {zx: 0 for zx in np.arange(-1, mrule_base.n_linguistic_variables()[key])}

            patterns_dss = {ix: {} for ix in range(len(mrule_base))}

        for ix, rule_base in enumerate(mrule_base):
            if len(rule_base) != 0:
                unique_patterns, patterns_ds, var_used = self.count_unique_patterns(rule_base)
                class_patterns[ix] = add_dicts(class_patterns[ix], unique_patterns)
                for key, value in class_vars.items():
                    class_vars[ix][key] = add_dicts(class_vars[ix][key], var_used[key])

                patterns_dss[ix] = concatenate_dicts(patterns_dss[ix], patterns_ds)
            

        return class_patterns, patterns_dss, class_vars
    

    def get_patterns_scores(self, n=30, n_gen=10, pop_size=10,test_size=0.33):
        '''
        Gets the patterns scores for the generated solutions.

        :param n: int. The number of solutions to generate.
        :param test_size: int defines train/test split
        :return class_patterns: dict[list]. The dictionary with the unique patterns for each class.
        :return patterns_dss: dict[list]. The dictionary with the dominance score of each pattern for each class.
        :return class_vars: dict[list]. The dictionary with the number of times each variable is used in the rule base for each class.
        :return accuracies: list. The list with the accuracies of the generated solutions.
        :return rule_bases: list. The list with the generated rule bases.
        '''
        rule_bases, accuracies = self.generate_solutions(n, n_gen=n_gen, pop_size=pop_size,test_size=test_size)
        self.n = n
        faults = 0
        for ix, mrule_base in enumerate(rule_bases):
            if len(mrule_base) != 0:
                if ix == 0:
                    class_patterns, patterns_dss, class_vars = self.count_unique_patterns_all_classes(mrule_base)
                else:
                    class_patterns, patterns_dss, class_vars = self.count_unique_patterns_all_classes(mrule_base, class_patterns, patterns_dss, class_vars)
            else:
                faults += 1
                print(f'No rules were generated for solution {ix}. Percentage of faulty solutions: {faults / n * 100}%')
            
        # Sort the patterns by the number of appearances
        for ix in range(len(class_patterns)):
            class_patterns[ix] = dict(sorted(class_patterns[ix].items(), key=lambda item: item[1], reverse=True))
            patterns_dss[ix] = dict(sorted(patterns_dss[ix].items(), key=lambda item: item[1], reverse=True))

        return class_patterns, patterns_dss, class_vars, accuracies, rule_bases
    

    def var_reports(self, class_vars: dict, antecedents, cutoff=10):
        '''
        Generates variable reports.
        
        :param class_vars: dict. The dictionary with the number of times each variable is used in the rule base for each class.
        :param antecedents: list. The list of antecedents.
        :param cutoff: int. The number of variables to show in the report.
        '''
        for jx in range(len(class_vars)):
            initiated = False
            sorted_class_ix = dict(sorted(class_vars[jx].items(), key=lambda item: item[1], reverse=True))
            for ix, key in enumerate(sorted_class_ix):
                if key != -1 and sorted_class_ix[key] > 0:
                    if ix > cutoff:
                        break
                    if not initiated:
                        print(f'Variable {antecedents[jx].name}')
                        initiated = True

                    print(f'{antecedents[jx][key].name} appears %.2f times per experiment.' % float(class_vars[jx][key] / self.n))
            if initiated:
                print()
                

    def text_report(self, class_patterns: dict, patterns_dss: dict, class_vars: dict, accuracies: list, rule_bases: list,
                    rule_cutoff: int = 5):
        '''
        Generates a text report for pattern stability.

        :param class_patterns: dict[list]. The dictionary with the unique patterns for each class.
        :param patterns_dss: dict[list]. The dictionary with the dominance score of each pattern for each class.
        :param class_vars: dict[list]. The dictionary with the number of times each variable is used in the rule base for each class.
        :param accuracies: list. The list with the accuracies of the generated solutions.
        :param rule_bases: list. The list with the generated rule bases.
        :param rule_cutoff: int. The number of rules to show in the report.
        '''
        consequents_names = self.classes_names
        print(f'Pattern stability report for {self.n} generated solutions')
        print('Average accuracy: %.2f\pm%.2f' % (np.mean(accuracies), np.std(accuracies)))
        print('-------------')

        for ix in range(len(class_patterns)):
            class_pattern_ix = class_patterns[ix]
            rules_array_format = [str_rule_as_list(key) for key in class_pattern_ix.keys()]
            patterns_dss_ix = patterns_dss[ix]
            class_vars_ix = class_vars[ix]

            print(f'Class {consequents_names[ix]}')
            print(f'Number of unique patterns: {len(class_pattern_ix)}')
            for jx, rule in enumerate(class_pattern_ix.keys()):
                if jx < rule_cutoff:
                    rule_print_format = rl.generate_rule_string(rules_array_format[jx], rule_bases[ix].antecedents)
                    print(f'Pattern {rule_print_format} appears in %.2f percent of the trials with a Dominance Score of {patterns_dss_ix[str(rule)]}' % float(class_pattern_ix[str(rule)] / self.n))
                else:
                    break
            print()
            self.var_reports(class_vars_ix, rule_bases[0].antecedents)
            print()
    

    def stability_report(self, n=10, n_gen=30, pop_size=20,test_size=0.33):
        '''
        Generates a stability report for pattern stabilization.

        :param n: int. The number of solutions to generate.
        :return text_report: str. The text report for pattern stability.
        '''
        class_patterns, patterns_dss, class_vars, accuracies, rule_bases = self.get_patterns_scores(n, n_gen=n_gen, pop_size=pop_size,test_size=test_size)
        self.class_patterns = class_patterns
        self.patterns_dss = patterns_dss
        self.class_vars = class_vars
        self.accuracies = accuracies
        self.rule_bases = rule_bases

        return self.text_report(class_patterns, patterns_dss, class_vars, accuracies, rule_bases)
    

    def pie_chart_basic(self, var_ix: int, class_ix: int):
        '''
        Generates a pie chart for the variable usage in the rule bases.
        
        :param var_ix: int. The index of the variable to analyze.
        :param class_ix: int. The index of the class to analyze.
        '''
        antecedents = self.rule_bases[0][class_ix].antecedents
        labels = []
        sizes = []
        var = self.class_vars[class_ix][var_ix]

        for key in var.keys():
            if key != -1 and var[key] > 0:
                labels.append(antecedents[key].name)
                sizes.append(var[key])

        fig1, ax1 = plt.subplots()
        ax1.pie(sizes, labels=labels, autopct='%1.1f%%',
                shadow=True, startangle=90)
        ax1.axis('equal')
        plt.show()

    
    def pie_chart_var(self, var_ix: int):
        '''
        Generates a pie chart for the variable usage in the rule bases.

        :param var_ix: int. The index of the variable to analyze.
        '''
        antecedents = self.rule_bases[0][0].antecedents
        colors = self.gen_colormap(antecedents)

        fig1, ax1 = plt.subplots(ncols=len(self.rule_bases[0]), nrows=1, figsize=(20, 10))
        fig1.suptitle(f'Variable {antecedents[var_ix].name} usage in the rulebases')

        for class_ix in range(len(self.rule_bases[0])):
            labels = []
            sizes = []
            var = self.class_vars[class_ix][var_ix]
            ax1[class_ix].set_title(f'Class {self.classes_names[class_ix]}')
            for key in var.keys():
                if key != -1 and var[key] > 0:
                    labels.append(antecedents[var_ix][key].name)
                    sizes.append(var[key])
        
            ax1[class_ix].pie(sizes, labels=labels, autopct='%1.1f%%', shadow=False, startangle=90, colors=[colors[v] for v in labels])
            ax1[class_ix].axis('equal')

        plt.show()

    
    def pie_chart_class(self, class_ix: int, var_list:list[int] = None):
        '''
        Generates a pie chart for the variable usage in the rule bases.

        :param class_ix: int. The index of the class to analyze.
        :param var_list: list[int]. The list of variables to analyze. If None (default) all variables will be analyzed.
        '''
        antecedents = self.rule_bases[0][0].antecedents
        colors = self.gen_colormap(antecedents)
        
        fig1, ax1 = plt.subplots(ncols=len(self.rule_bases[0]), nrows=1, figsize=(20, 10))
        fig1.suptitle(f'Class {self.classes_names[class_ix]} variable usage in the rulebases')

        for var_ix in range(len(self.rule_bases[0])):
            if (var_list is not None and var_ix in var_list) or var_list is None:
                labels = []
                sizes = []
                var = self.class_vars[class_ix][var_ix]
                ax1[var_ix].set_title(f'Variable {antecedents[var_ix].name}')
                for key in var.keys():
                    if key != -1 and var[key] > 0:
                        labels.append(antecedents[var_ix][key].name)
                        sizes.append(var[key])
            
                ax1[var_ix].pie(sizes, labels=labels, autopct='%1.1f%%', shadow=False, startangle=90, colors=[colors[v] for v in labels])
                ax1[var_ix].axis('equal')

        plt.show()


    def gen_colormap(self, antecedents):
        '''
        Generates a colormap for the special cases of 2 and 3 linguistic variables.

        :param antecedents: list. The list of antecedents.
        '''
        largest_vl_ix = np.argmax(self.n_linguist_variables)
        largest_vl_n = self.n_linguist_variables[largest_vl_ix]

        # Note: red and yellow has been softened to avoid eye strain, that's why there are colors specified with hexadecimals
        if largest_vl_n == 2: # There is the special case of low/high
            colors = { label: color  for label, color in zip([antecedent.name for antecedent in antecedents[largest_vl_ix]], ['#FA8072', 'Green'])} 
        elif largest_vl_n == 3: # There is the special case of low/medium/high
            colors = { label: color  for label, color in zip([antecedent.name for antecedent in antecedents[largest_vl_ix]], ['#FA8072', '#EEE8AA', 'Green'])}
        else:
            colormap_custom = list(set([colormaps['coolwarm'](a) for a in len(largest_vl_n)]))
            
            colors = { label: color  for label, color in zip([antecedent.name for antecedent in antecedents], colormap_custom)}
        return colors
            



"""
Functions that contain some general functions to eval already fitted fuzzy rule based models.
It can also be used to visualize rules and fuzzy partitions.
"""
import numpy as np
import pandas as pd
import sklearn.metrics as metrics

try:
      from . import evolutionary_fit as evf
      from . import vis_rules
except ImportError:
      import evolutionary_fit as evf
      import vis_rules
    
    
def eval_fuzzy_model(fl_classifier: evf.BaseFuzzyRulesClassifier, X_train:np.array, y_train:np.array, X_test:np.array, y_test:np.array, plot_rules=False,print_rules:bool=True, plot_partitions:bool=False, return_rules:bool=True, bootstrap_results_print:bool=True) -> str:
      '''
      Take a look at the FuzzyEvaluator class for this function documentation.
      '''
      fuzzy_evaluator = FuzzyEvaluator(fl_classifier)
      res = fuzzy_evaluator.eval_fuzzy_model(X_train, y_train, X_test, y_test, 
                        plot_rules=plot_rules, print_rules=print_rules, plot_partitions=plot_partitions, return_rules=return_rules, bootstrap_results_print=bootstrap_results_print)
      

      return res

class FuzzyEvaluator():
    '''
    Takes a model and associated data and permits rule evaluation
    '''
    def __init__(self,fl_classifier: evf.BaseFuzzyRulesClassifier):
            '''
            :param fl_classifier: Fuzzy rule based model
            '''
            self.fl_classifier = fl_classifier


    
    def predict(self,X: np.array) -> np.array:
        # Predict y for given X for use in metric evaluation
        return self.fl_classifier.predict(X)
    

    def get_metric(self,metric:str,X_true:np.array,y_true:np.array,**kwargs) -> float:
        '''
        :param metric: named metric in string format available in sklearn library for evaluation
        :param X_true: np.array of X values for prediction
        :param y_true: np.array of true class outcomes for X values
        :param **kwargs: additional arguments for different sklearn.metrics functions
        '''
        #Get y predictions
        y_pred = self.predict(X_true)
        y_true = np.array(y_true)
        #Convert str classes to numbers in corresponding class if necessary
        unique_classes = self.fl_classifier.classes_names

        if isinstance(y_true[0],str):
            y_true = np.array([list[unique_classes].index(str(y)) for y in y_true])
      
        #Find metrics requested in sklearn library, if not found 
        try:
             # Get the metric function dynamically from sklearn.metrics
              metric_function = getattr(metrics, metric)
             # Call the metric function with y_true, y_pred, and any additional keyword arguments
              return metric_function(y_true, y_pred, **kwargs)
        except AttributeError:
              return f"Metric '{metric}' not found in sklearn.metrics."
        except TypeError:
              return f"Invalid arguments passed for the metric '{metric}'."
        
        
    def eval_fuzzy_model(self,X_train: np.array, y_train: np.array,X_test: np.array, y_test: np.array, 
                         plot_rules=True, print_rules=True, plot_partitions=True, 
                     return_rules=False, print_accuracy=True, print_matthew=True, export_path:str=None, bootstrap_results_print:bool=True) -> None:
      '''
      Function that evaluates a fuzzy rule based model. It also plots the rules and the fuzzy partitions.

      :param X_train: Training data.
      :param y_train: Training labels.
      :param X_test: Test data.
      :param y_test: Test labels.
      :param plot_rules: If True, it plots the rules.
      :param print_rules: If True, it prints the rules.
      :param plot_partitions: If True, it plots the fuzzy partitions.
      :return: None
      '''
      # Get the unique classes from the classifier
      unique_classes = self.fl_classifier.classes_names

      # Convert the names from the labels to the corresponding class if necessary
      if isinstance(y_train[0], str):
            y_train = np.array([list(unique_classes).index(str(y)) for y in y_train])
            y_test = np.array([list(unique_classes).index(str(y)) for y in y_test])
      
      if print_accuracy:
            print('------------')
            print('ACCURACY')
            print('Train performance: ' +
                  str(self.get_metric('accuracy_score',X_train,y_train)))
            print('Test performance: ' +
                  str(self.get_metric('accuracy_score',X_test,y_test)))
            print('------------')
      if print_matthew:
            print('MATTHEW CORRCOEF')
            print('Train performance: ' +
                  str(self.get_metric('matthews_corrcoef',X_train,y_train))
                  )
            print('Test performance: ' +
                  str(self.get_metric('matthews_corrcoef',X_test,y_test))
                  )
            print('------------')


      if print_rules or return_rules:
            res = self.fl_classifier.print_rules(True, bootstrap_results=bootstrap_results_print)

            if print_rules:
                  print(res)

      if plot_partitions:
            self.fl_classifier.plot_fuzzy_variables()

      if plot_rules:
            vis_rules.visualize_rulebase(self.fl_classifier.rule_base, export_path=export_path)
            
      if return_rules:
            return res